/*
 * JODG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jodg;


import java.util.*;

import java.awt.BasicStroke;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.Paint;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.RenderingHints.Key;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.font.FontRenderContext;
import java.awt.font.GlyphVector;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.BufferedImageOp;
import java.awt.image.ImageObserver;
import java.awt.image.RenderedImage;
import java.awt.image.renderable.RenderableImage;
import java.text.AttributedCharacterIterator;
import java.text.CharacterIterator;



/**
 * The ODGGraphics class is the main class of the ODGGraphics library. It extends the standard
 * Java Graphics and Graphics2D classes, reimplementing their methods to generate ODG objects
 * which can then generate ODG output for use with CAD programs. A ODGGraphics object is always 
 * associated with an ODGDocument instance, and is obtained from the ODGDocument through a call to
 * the getGraphics() method. The typical workflow is as follows:
 * 
 * <pre>
 * 
 * // Create an ODG document and get its associated ODGGraphics instance
 * ODGDocument odgDocument = new ODGDocument(); 
 * ODGGraphics graphics = odgDocument.getGraphics(); 
 * 
 * // Do drawing commands as on any other Graphics. If you have a paint(Graphics) method, 
 * // you can just use it with the ODGGraphics instance since it's a subclass of Graphics. 
 * graphics.setColor(Color.RED);  
 * graphics.setStroke(new BasicStroke(3));
 * graphics.drawLine(0, 0, 1000, 500); 
 * graphics.drawRect(1000, 500, 150, 150); 
 * graphics.drawRoundRect(20, 200, 130, 100, 20, 10); 
 * 
 * // Get the flat ODG output as a string - it's just XML text - and  save  in a file for use 
 * // with LibreOffice Draw or any other draw package that accepts flat ODG (.fodg) files
 * String odgText = odgDocument.toODGString(); 
 * String filePath = "path/to/file.fodg"; 
 * FileWriter fileWriter = new FileWriter(filePath); 
 * fileWriter.write(odgText); 
 * fileWriter.flush(); 
 * fileWriter.close();
 * 
 * </pre>
 * 
 * @author jsevy
 */
public class ODGGraphics extends Graphics2D
{

    private ODGDocument odgDocument;
    
    // class access; used to indicate if circles and circular arcs should be generated,
    // or just more general elliptical arcs
    boolean useCircles;
    
    private Color color;
    private Color backColor;
    private PaintMode paintMode;
    private Font font;
    private BasicStroke stroke;  // pen parameters
    
    private AffineTransform javaTransform;
    
    private enum PaintMode
    {
        PAINT_MODE,
        XOR_MODE
    };
    
    
    /**
     * Constructs a new ODGGraphics object. Since an ODGGraphics object is
     * always associated with an ODGDocument instance, applications
     * should not call this constructor directly, but rather use the getGraphics() method 
     * of an ODGDocument.
     * 
     * @param   odgDocument     The associated ODGDocument instance
     */
    public ODGGraphics(ODGDocument odgDocument) 
    {
        this.odgDocument = odgDocument;
        
        useCircles = true;
        
        color = Color.BLACK;
        backColor = Color.WHITE;
        paintMode = PaintMode.PAINT_MODE;
        font = new Font(Font.SANS_SERIF, Font.PLAIN, 12);
        stroke = new BasicStroke();
        
        // user-defined Java transform - initialize to identity matrix
        javaTransform = new AffineTransform();
        
    }

    /**
     * Creates a new ODGGraphics object that is a copy of this ODGGraphics object.
     * Since an ODGGraphics object is always associated with an ODGDocument instance, applications
     * should not call this constructor directly, but rather use the getGraphics() method 
     * of an ODGDocument.
     * @return     a new ODGGraphics context that is a copy of
     *                       this graphics context.
     */
    public ODGGraphics create()
    {
        ODGGraphics newGraphics = new ODGGraphics(this.odgDocument);
        
        newGraphics.color = this.color;
        newGraphics.backColor = this.backColor;
        newGraphics.paintMode = this.paintMode;
        newGraphics.font = this.font;
        newGraphics.stroke = this.stroke;
        
        newGraphics.javaTransform = new AffineTransform(this.javaTransform);
        
        return newGraphics;
    }

    /**
     * Creates a new ODGGraphics object based on this ODGGraphics object, but with a new translation.
     * Since an ODGGraphics object is always associated with an ODGDocument instance, applications
     * should not call this constructor directly, but rather use the getGraphics() method 
     * of an ODGDocument.
     */
    public ODGGraphics create(int x, int y, int width, int height) {
        ODGGraphics graphics = create();
        if (graphics == null) return null;
        graphics.translate(x, y);
        return graphics;
    }
    
    
    
    /**
     * Gets this graphic's foreground color.
     * @return    the foreground color
     */
    public Color getColor()
    {
        return color;
    }

    /**
     * Sets this graphics foreground color to the specified
     * color. All subsequent graphics operations using this graphics
     * objects use this new color.
     * @param     color   the new rendering color.
     */
    public void setColor(Color color)
    {
        this.color = color;
    }

    /**
     * Sets the paint mode of this graphics context to overwrite the
     * destination with this graphics context's current color. Note that
     * this is the only mode supported by ODGGraphics.
     */
    public void setPaintMode()
    {
        paintMode = PaintMode.PAINT_MODE;
    }

    /**
     * Currently does nothing, as only PAINT_MODE is supported
     */
    public void setXORMode(Color xorColor)
    {
        paintMode = PaintMode.XOR_MODE;
    }

    /**
     * Gets the current font.
     * @return    this graphic's current font.
     */
    public Font getFont()
    {
        return font;
    }

    /**
     * Sets this graphics context's font to the specified font.
     * All subsequent text operations using this graphics context
     * use this font. A null argument is silently ignored.
     * @param  font   the new font.
    */
    public void setFont(Font font)
    {
        if (font != null)
            this.font = font;
    }

    /**
     * Gets the font metrics of the current font.
     * @return    the font metrics of this graphics
     *                    context's current font.
     */
    public FontMetrics getFontMetrics() 
    {
        return getFontMetrics(getFont());
    }

    /**
     * Gets the font metrics for the specified font.
     * @return    the font metrics for the specified font.
     * @param     f the specified font
     */
    public FontMetrics getFontMetrics(Font f)
    {
        // get the system FontMetrics for the current Font; will not match precisely the ODG font
        Canvas c = new Canvas();
        return c.getFontMetrics(f);
    }


    /**
     * Returns null, as clipping not supported.
     * @return      null
     */
    public Rectangle getClipBounds()
    {
        return null;
    }

    /**
     * Does nothing, as clipping not supported
     */
    public void clipRect(int x, int y, int width, int height)
    {
        // do nothing
    }

    /**
     * Does nothing, as clipping not supported
     */
    public void setClip(int x, int y, int width, int height)
    {
        // do nothing
    }

    /**
     * Returns null, as clipping not supported
     * @return      null
     */
    public Shape getClip()
    {
        return null;
    }

    /**
     * Does nothing, as clipping not supported
     */
    public void setClip(Shape clip)
    {
        // do nothing
    }

    /**
     * Throws UnsupportedOperationException, as copyArea not supported.
     */
    public void copyArea(int x, int y, int width, int height, int dx, int dy)
    {
        throw new UnsupportedOperationException("copyArea not supported");
    }

    /**
     * Draws a line, using the current color, between the points
     * <code>(x1,&nbsp;y1)</code> and <code>(x2,&nbsp;y2)</code>
     * in this graphics context's coordinate system.
     * @param   x1  the first point's <i>x</i> coordinate.
     * @param   y1  the first point's <i>y</i> coordinate.
     * @param   x2  the second point's <i>x</i> coordinate.
     * @param   y2  the second point's <i>y</i> coordinate.
     */
    public void drawLine(int x1, int y1, int x2, int y2)
    {
        this.drawLine((double)x1, (double)y1, (double)x2, (double)y2);
    }
    
    /**
     * Draws a line, using the current color, between the points
     * <code>(x1,&nbsp;y1)</code> and <code>(x2,&nbsp;y2)</code>
     * in this graphics context's coordinate system.
     * @param   x1  the first point's <i>x</i> coordinate.
     * @param   y1  the first point's <i>y</i> coordinate.
     * @param   x2  the second point's <i>x</i> coordinate.
     * @param   y2  the second point's <i>y</i> coordinate.
     */
    public void drawLine(double x1, double y1, double x2, double y2)
    {
        ODGLine line = createODGLine(x1, y1, x2, y2);
        odgDocument.addEntity(line);
    }
    
    /**
     * Creates an ODGLine object corresponding to the line between
     * the supplied endpoints.
     * @param   x1  the first point's <i>x</i> coordinate.
     * @param   y1  the first point's <i>y</i> coordinate.
     * @param   x2  the second point's <i>x</i> coordinate.
     * @param   y2  the second point's <i>y</i> coordinate.
     * @return  corresponding ODGLine Object
     */
    private ODGLine createODGLine(double x1, double y1, double x2, double y2)
    {
        ODGGraphicsStyle style = odgDocument.getGraphicsStyle(this, true, false);
        return new ODGLine(new RealPoint(x1, y1, 0), new RealPoint(x2, y2, 0), style, javaTransform);
    }

    /**
     * Fills the specified rectangle.
     * The left and right edges of the rectangle are at
     * <code>x</code> and <code>x&nbsp;+&nbsp;width&nbsp;-&nbsp;1</code>.
     * The top and bottom edges are at
     * <code>y</code> and <code>y&nbsp;+&nbsp;height&nbsp;-&nbsp;1</code>.
     * The rectangle is filled using the graphics context's current color.
     * @param         x   the <i>x</i> coordinate
     *                         of the rectangle to be filled.
     * @param         y   the <i>y</i> coordinate
     *                         of the rectangle to be filled.
     * @param         width   the width of the rectangle to be filled.
     * @param         height   the height of the rectangle to be filled.
     */
    public void fillRect(int x, int y, int width, int height)
    {
        this.fillRect((double)x, (double)y, (double)width, (double)height);
    }
    
    
    /**
     * Fills the specified rectangle.
     * The left and right edges of the rectangle are at
     * <code>x</code> and <code>x&nbsp;+&nbsp;width&nbsp;-&nbsp;1</code>.
     * The top and bottom edges are at
     * <code>y</code> and <code>y&nbsp;+&nbsp;height&nbsp;-&nbsp;1</code>.
     * The rectangle is filled using the graphics context's current color.
     * @param         x   the <i>x</i> coordinate
     *                         of the rectangle to be filled.
     * @param         y   the <i>y</i> coordinate
     *                         of the rectangle to be filled.
     * @param         width   the width of the rectangle to be filled.
     * @param         height   the height of the rectangle to be filled.
     */
    public void fillRect(double x, double y, double width, double height)
    {
        drawFillRect(x, y, width, height, true);
    }

    /**
     * Draws the outline of the specified rectangle.
     * The left and right edges of the rectangle are at
     * <code>x</code> and <code>x&nbsp;+&nbsp;width</code>.
     * The top and bottom edges are at
     * <code>y</code> and <code>y&nbsp;+&nbsp;height</code>.
     * The rectangle is drawn using the graphics context's current color.
     * @param         x   the <i>x</i> coordinate
     *                         of the rectangle to be drawn.
     * @param         y   the <i>y</i> coordinate
     *                         of the rectangle to be drawn.
     * @param         width   the width of the rectangle to be drawn.
     * @param         height   the height of the rectangle to be drawn.
     */
    public void drawRect(int x, int y, int width, int height)
    {
        this.drawRect((double)x, (double)y, (double)width, (double)height);
    }
    
    
    /**
     * Draws the outline of the specified rectangle.
     * The left and right edges of the rectangle are at
     * <code>x</code> and <code>x&nbsp;+&nbsp;width</code>.
     * The top and bottom edges are at
     * <code>y</code> and <code>y&nbsp;+&nbsp;height</code>.
     * The rectangle is drawn using the graphics context's current color.
     * @param         x   the <i>x</i> coordinate
     *                         of the rectangle to be drawn.
     * @param         y   the <i>y</i> coordinate
     *                         of the rectangle to be drawn.
     * @param         width   the width of the rectangle to be drawn.
     * @param         height   the height of the rectangle to be drawn.
     */
    public void drawRect(double x, double y, double width, double height)
    {
        drawFillRect(x, y, width, height, false);
    }
    
    
    private void drawFillRect(double x, double y, double width, double height, boolean filled)
    {
        ODGGraphicsStyle style;
        if (filled)
            style = odgDocument.getGraphicsStyle(this, false, true);
        else
            style = odgDocument.getGraphicsStyle(this, true, false);
        
        // now see if we're pre-applying the transform to work around LibreOffice bugs in handling
        // shear and scaling transformations - consider some different cases
        if ((!ODGEntity.preApplyTransform) || (javaTransform.isIdentity()))
        {
            // either way, just insert rectangle as is
            ODGRectangle rectangle = new ODGRectangle(x, y, width, height, 0 /*corner radius*/, style, javaTransform);
            odgDocument.addEntity(rectangle);
        }
        else
        {
            // we're applying the transform and the transform isn't the identity; 
            // in this case our rectangle may become a polyline
            double[] cornerPoints = {x, y, x+width, y, x+width, y+height, x, y+height};
            javaTransform.transform(cornerPoints, 0, cornerPoints, 0, 4);
            
            // see if it's still a rectangle; check sides for perpendicularity
            RealPoint topVector = new RealPoint(cornerPoints[2]-cornerPoints[0], cornerPoints[3]-cornerPoints[1], 0);
            RealPoint sideVector = new RealPoint(cornerPoints[4]-cornerPoints[2], cornerPoints[5]-cornerPoints[3], 0);
            if ((RealPoint.dotProduct(topVector, sideVector) == 0) && ((topVector.x == 0) || (topVector.y == 0)))
            {
                // still a rectangle, and oriented with sides parallel to the axes; can still represent as ODGRectangle
                // find dimensions
                double[] xCoords = {cornerPoints[0], cornerPoints[2], cornerPoints[4], cornerPoints[6]};
                double[] yCoords = {cornerPoints[1], cornerPoints[3], cornerPoints[5], cornerPoints[7]};
                
                // get crafty
                Arrays.sort(xCoords);
                Arrays.sort(yCoords);
                
                double newWidth = Math.abs(xCoords[2]-xCoords[0]);
                double newHeight = Math.abs(yCoords[2]-yCoords[0]);
                double newLeft = xCoords[0];
                double newTop = yCoords[0];
                
                ODGRectangle rectangle = new ODGRectangle(newLeft, newTop, newWidth, newHeight, 0 /*corner radius*/, style, new AffineTransform());
                odgDocument.addEntity(rectangle);
                
            }
            else
            {
                // just represent as a polyline; use the standard polyline draw call with the original points
                double[] xCoords = {x, x+width, x+width, x};
                double[] yCoords = {y, y, y+height, y+height};
                if (filled)
                {
                    fillPolygon(xCoords, yCoords, 4);
                }
                else
                {
                    drawPolygon(xCoords, yCoords, 4);
                }
            }
            
        }
    }
    

    /**
     * Clears the specified rectangle by filling it with the background
     * color.
     * @param       x the <i>x</i> coordinate of the rectangle to clear.
     * @param       y the <i>y</i> coordinate of the rectangle to clear.
     * @param       width the width of the rectangle to clear.
     * @param       height the height of the rectangle to clear.
     */
    public void clearRect(int x, int y, int width, int height)
    {
        this.clearRect((double)x, (double)y, (double)width, (double)height);
    }
    
    
    /**
     * Clears the specified rectangle by filling it with the background
     * color.
     * @param       x the <i>x</i> coordinate of the rectangle to clear.
     * @param       y the <i>y</i> coordinate of the rectangle to clear.
     * @param       width the width of the rectangle to clear.
     * @param       height the height of the rectangle to clear.
     */
    public void clearRect(double x, double y, double width, double height)
    {
        // Create Shape and use general Shape routine
        Rectangle2D rect = new Rectangle2D.Double(x, y, width, height);
        clear(rect);
    }

    /**
     * Draws an outlined round-cornered rectangle using this graphics
     * context's current color. The left and right edges of the rectangle
     * are at <code>x</code> and <code>x&nbsp;+&nbsp;width</code>,
     * respectively. The top and bottom edges of the rectangle are at
     * <code>y</code> and <code>y&nbsp;+&nbsp;height</code>.
     * @param      x the <i>x</i> coordinate of the rectangle to be drawn.
     * @param      y the <i>y</i> coordinate of the rectangle to be drawn.
     * @param      width the width of the rectangle to be drawn.
     * @param      height the height of the rectangle to be drawn.
     * @param      arcWidth the horizontal diameter of the arc
     *                    at the four corners.
     * @param      arcHeight the vertical diameter of the arc
     *                    at the four corners.
     */
    public void drawRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight)
    {
        this.drawRoundRect((double)x, (double)y, (double)width, (double)height, (double)arcWidth, (double)arcHeight);        
    }
    
    
    
    /**
     * Draws an outlined round-cornered rectangle using this graphics
     * context's current color. The left and right edges of the rectangle
     * are at <code>x</code> and <code>x&nbsp;+&nbsp;width</code>,
     * respectively. The top and bottom edges of the rectangle are at
     * <code>y</code> and <code>y&nbsp;+&nbsp;height</code>.
     * @param      x the <i>x</i> coordinate of the rectangle to be drawn.
     * @param      y the <i>y</i> coordinate of the rectangle to be drawn.
     * @param      width the width of the rectangle to be drawn.
     * @param      height the height of the rectangle to be drawn.
     * @param      arcWidth the horizontal diameter of the arc
     *                    at the four corners.
     * @param      arcHeight the vertical diameter of the arc
     *                    at the four corners.
     */
    public void drawRoundRect(double x, double y, double width, double height, double arcWidth, double arcHeight)
    {
        
        // see if arc width and height are same, and if we're not pre-applying the transform or it's the identity
        if ((arcWidth == arcHeight) && ((!ODGEntity.preApplyTransform) || (javaTransform.isIdentity())))
        {
            // can use standard ODG rectangle with rounded corners  
            ODGGraphicsStyle style = odgDocument.getGraphicsStyle(this, true, false); 
            ODGRectangle rectangle = new ODGRectangle(x, y, width, height, arcWidth, style, javaTransform);
            odgDocument.addEntity(rectangle);
        }
        else
        {
            // just create as series of lines and elliptical arcs
            double left = x;
            double right = x + width;
            double top = y;
            double bottom = y + height;
            
            double rx = arcWidth/2;
            double ry = arcHeight/2;
            
            // ick: draw ovals in corners connected by lines, clockwise from top left
            // note that the subroutines will take care of transforming these as needed
            
            // top line and top-right oval segment
            drawLine(left + rx, top, right - rx, top);
            drawArc(right - 2 * rx, top, arcWidth, arcHeight, 90, -90);
            
            // right line and bottom-right oval segment
            drawLine(right, top+ry, right, bottom - ry);
            drawArc(right - 2 * rx, bottom - 2 * ry, arcWidth, arcHeight, 0, -90);
            
            // bottom line and bottom-left oval segment
            drawLine(right-rx, bottom, left + rx, bottom);
            drawArc(left, bottom - 2 * ry, arcWidth, arcHeight, 270, -90);
            
            // left line and top-left arc
            drawLine(left, bottom-ry,left, top + ry);
            drawArc(left, top, arcWidth, arcHeight, 180, -90);
        }
    }
    
    
   
    /**
     * Fills the specified rounded corner rectangle with the current color.
     * The left and right edges of the rectangle
     * are at <code>x</code> and <code>x&nbsp;+&nbsp;width&nbsp;-&nbsp;1</code>,
     * respectively. The top and bottom edges of the rectangle are at
     * <code>y</code> and <code>y&nbsp;+&nbsp;height&nbsp;-&nbsp;1</code>.
     * @param       x the <i>x</i> coordinate of the rectangle to be filled.
     * @param       y the <i>y</i> coordinate of the rectangle to be filled.
     * @param       width the width of the rectangle to be filled.
     * @param       height the height of the rectangle to be filled.
     * @param       arcWidth the horizontal diameter
     *                     of the arc at the four corners.
     * @param       arcHeight the vertical diameter
     *                     of the arc at the four corners.
     */
    public void fillRoundRect(int x, int y, int width, int height, int arcWidth, int arcHeight)
    {
        this.fillRoundRect((double)x, (double)y, (double)width, (double)height, (double)arcWidth, (double)arcHeight);
    }
    
    
    /**
     * Fills the specified rounded corner rectangle with the current color.
     * The left and right edges of the rectangle
     * are at <code>x</code> and <code>x&nbsp;+&nbsp;width&nbsp;-&nbsp;1</code>,
     * respectively. The top and bottom edges of the rectangle are at
     * <code>y</code> and <code>y&nbsp;+&nbsp;height&nbsp;-&nbsp;1</code>.
     * @param       x the <i>x</i> coordinate of the rectangle to be filled.
     * @param       y the <i>y</i> coordinate of the rectangle to be filled.
     * @param       width the width of the rectangle to be filled.
     * @param       height the height of the rectangle to be filled.
     * @param       arcWidth the horizontal diameter
     *                     of the arc at the four corners.
     * @param       arcHeight the vertical diameter
     *                     of the arc at the four corners.
     */
    public void fillRoundRect(double x, double y, double width, double height, double arcWidth, double arcHeight)
    {
        
        // see if arc width and height are same, and if we're not pre-applying the transform or it's the identity
        if ((arcWidth == arcHeight) && ((!ODGEntity.preApplyTransform) || (javaTransform.isIdentity())))
        {
            // can use standard ODG rectangle with rounded corners 
            ODGGraphicsStyle style = odgDocument.getGraphicsStyle(this, false, true); 
            ODGRectangle rectangle = new ODGRectangle(x, y, width, height, arcWidth, style, javaTransform);
            odgDocument.addEntity(rectangle);
        }
        else
        {
            // Create Shape and use general Shape routine
            RoundRectangle2D rect = new RoundRectangle2D.Double(x, y, width, height, arcWidth, arcHeight);
            fill(rect);
        }
    }
    
    
    

    /**
     * Draws a 3-D highlighted outline of the specified rectangle.
     * The edges of the rectangle are highlighted so that they
     * appear to be beveled and lit from the upper left corner.
     * <p>
     * The colors used for the highlighting effect are determined
     * based on the current color.
     * The resulting rectangle covers an area that is
     * <code>width&nbsp;+&nbsp;1</code> pixels wide
     * by <code>height&nbsp;+&nbsp;1</code> pixels tall.
     * @param       x the <i>x</i> coordinate of the rectangle to be drawn.
     * @param       y the <i>y</i> coordinate of the rectangle to be drawn.
     * @param       width the width of the rectangle to be drawn.
     * @param       height the height of the rectangle to be drawn.
     * @param       raised a boolean that determines whether the rectangle
     *                      appears to be raised above the surface
     *                      or sunk into the surface.
     */
    public void draw3DRect(int x, int y, int width, int height,
                           boolean raised) {
        super.draw3DRect(x, y, width, height, raised);
    }

    /**
     * Paints a 3-D highlighted rectangle filled with the current color.
     * The edges of the rectangle will be highlighted so that it appears
     * as if the edges were beveled and lit from the upper left corner.
     * The colors used for the highlighting effect will be determined from
     * the current color.
     * @param       x the <i>x</i> coordinate of the rectangle to be filled.
     * @param       y the <i>y</i> coordinate of the rectangle to be filled.
     * @param       width the width of the rectangle to be filled.
     * @param       height the height of the rectangle to be filled.
     * @param       raised a boolean value that determines whether the
     *                      rectangle appears to be raised above the surface
     *                      or etched into the surface.
     * @see         java.awt.Graphics#draw3DRect
     */
    public void fill3DRect(int x, int y, int width, int height,
                           boolean raised) {
        super.fill3DRect(x, y, width, height, raised);
    }

    /**
     * Draws the outline of an oval.
     * The result is a circle or ellipse that fits within the
     * rectangle specified by the <code>x</code>, <code>y</code>,
     * <code>width</code>, and <code>height</code> arguments.
     * @param       x the <i>x</i> coordinate of the upper left
     *                     corner of the oval to be drawn.
     * @param       y the <i>y</i> coordinate of the upper left
     *                     corner of the oval to be drawn.
     * @param       width the width of the oval to be drawn.
     * @param       height the height of the oval to be drawn.
     */
    public void drawOval (int x, int y, int width, int height)
    {
        this.drawOval((double)x, (double)y, (double)width, (double)height);
    }
    
    
    /**
     * Draws the outline of an oval.
     * The result is a circle or ellipse that fits within the
     * rectangle specified by the <code>x</code>, <code>y</code>,
     * <code>width</code>, and <code>height</code> arguments.
     * @param       x the <i>x</i> coordinate of the upper left
     *                     corner of the oval to be drawn.
     * @param       y the <i>y</i> coordinate of the upper left
     *                     corner of the oval to be drawn.
     * @param       width the width of the oval to be drawn.
     * @param       height the height of the oval to be drawn.
     */
    public void drawOval (double x, double y, double width, double height)
    {
        drawArc(x, y, width, height, 0, 360);
    }

    
    /**
     * Fills an oval bounded by the specified rectangle with the
     * current color.
     * @param       x the <i>x</i> coordinate of the upper left corner
     *                     of the oval to be filled.
     * @param       y the <i>y</i> coordinate of the upper left corner
     *                     of the oval to be filled.
     * @param       width the width of the oval to be filled.
     * @param       height the height of the oval to be filled.
     */
    public void fillOval(int x, int y, int width, int height)
    {
        this.fillOval((double)x, (double)y, (double)width, (double)height);
    }
    
    
    /**
     * Fills an oval bounded by the specified rectangle with the
     * current color.
     * @param       x the <i>x</i> coordinate of the upper left corner
     *                     of the oval to be filled.
     * @param       y the <i>y</i> coordinate of the upper left corner
     *                     of the oval to be filled.
     * @param       width the width of the oval to be filled.
     * @param       height the height of the oval to be filled.
     */
    public void fillOval(double x, double y, double width, double height)
    {
        fillArc(x, y, width, height, 0, 360);
    }

    
    /**
     * Draws the outline of a circular or elliptical arc
     * covering the specified rectangle.
     * <p>
     * The resulting arc begins at <code>startAngle</code> and extends
     * for <code>arcAngle</code> degrees, using the current color.
     * Angles are interpreted such that 0&nbsp;degrees
     * is at the 3&nbsp;o'clock position.
     * A positive value indicates a counter-clockwise rotation
     * while a negative value indicates a clockwise rotation.
     * <p>
     * The center of the arc is the center of the rectangle whose origin
     * is (<i>x</i>,&nbsp;<i>y</i>) and whose size is specified by the
     * <code>width</code> and <code>height</code> arguments.
     * <p>
     * The angles are specified relative to the non-square extents of
     * the bounding rectangle such that 45 degrees always falls on the
     * line from the center of the ellipse to the upper right corner of
     * the bounding rectangle. As a result, if the bounding rectangle is
     * noticeably longer in one axis than the other, the angles to the
     * start and end of the arc segment will be skewed farther along the
     * longer axis of the bounds.
     * @param        x the <i>x</i> coordinate of the
     *                    upper-left corner of the arc to be drawn.
     * @param        y the <i>y</i>  coordinate of the
     *                    upper-left corner of the arc to be drawn.
     * @param        width the width of the arc to be drawn.
     * @param        height the height of the arc to be drawn.
     * @param        startAngle the beginning angle, in degrees
     * @param        sweepAngle the angular extent of the arc, in degrees,
     *                    relative to the start angle.
     */
    public void drawArc(int x, int y, int width, int height, int startAngle, int sweepAngle)
    {
        this.drawArc((double)x, (double)y, (double)width, (double)height, (double)startAngle, (double)sweepAngle);
    }
    
    
    /**
     * Draws the outline of a circular or elliptical arc
     * covering the specified rectangle.
     * <p>
     * The resulting arc begins at <code>startAngle</code> and extends
     * for <code>arcAngle</code> degrees, using the current color.
     * Angles are interpreted such that 0&nbsp;degrees
     * is at the 3&nbsp;o'clock position.
     * A positive value indicates a counter-clockwise rotation
     * while a negative value indicates a clockwise rotation.
     * <p>
     * The center of the arc is the center of the rectangle whose origin
     * is (<i>x</i>,&nbsp;<i>y</i>) and whose size is specified by the
     * <code>width</code> and <code>height</code> arguments.
     * <p>
     * The angles are specified relative to the non-square extents of
     * the bounding rectangle such that 45 degrees always falls on the
     * line from the center of the ellipse to the upper right corner of
     * the bounding rectangle. As a result, if the bounding rectangle is
     * noticeably longer in one axis than the other, the angles to the
     * start and end of the arc segment will be skewed farther along the
     * longer axis of the bounds.
     * @param        x the <i>x</i> coordinate of the
     *                    upper-left corner of the arc to be drawn.
     * @param        y the <i>y</i>  coordinate of the
     *                    upper-left corner of the arc to be drawn.
     * @param        width the width of the arc to be drawn.
     * @param        height the height of the arc to be drawn.
     * @param        startAngle the beginning angle, in degrees
     * @param        sweepAngle the angular extent of the arc, in degrees,
     *                    relative to the start angle.
     */
    public void drawArc(double x, double y, double width, double height, double startAngle, double sweepAngle)
    {
        this.drawFillArc(x, y, width, height, startAngle, sweepAngle, false);
    }
    
    

    /**
     * Fills a circular or elliptical arc wedge covering the specified rectangle.
     * <p>
     * The resulting arc begins at <code>startAngle</code> and extends
     * for <code>arcAngle</code> degrees.
     * Angles are interpreted such that 0&nbsp;degrees
     * is at the 3&nbsp;o'clock position.
     * A positive value indicates a counter-clockwise rotation
     * while a negative value indicates a clockwise rotation.
     * <p>
     * The center of the arc is the center of the rectangle whose origin
     * is (<i>x</i>,&nbsp;<i>y</i>) and whose size is specified by the
     * <code>width</code> and <code>height</code> arguments.
     * <p>
     * The angles are specified relative to the non-square extents of
     * the bounding rectangle such that 45 degrees always falls on the
     * line from the center of the ellipse to the upper right corner of
     * the bounding rectangle. As a result, if the bounding rectangle is
     * noticeably longer in one axis than the other, the angles to the
     * start and end of the arc segment will be skewed farther along the
     * longer axis of the bounds.
     * @param        x the <i>x</i> coordinate of the
     *                    upper-left corner of the arc to be filled.
     * @param        y the <i>y</i>  coordinate of the
     *                    upper-left corner of the arc to be filled.
     * @param        width the width of the arc to be filled.
     * @param        height the height of the arc to be filled.
     * @param        startAngle the beginning angle.
     * @param        arcAngle the angular extent of the arc,
     *                    relative to the start angle.
     */
    public void fillArc(int x, int y, int width, int height, int startAngle, int arcAngle)
    {
        this.fillArc((double)x, (double)y, (double)width, (double)height, (double)startAngle, (double)arcAngle);
    }
    
    
    
    /**
     * Fills a circular or elliptical arc wedge covering the specified rectangle.
     * <p>
     * The resulting arc begins at <code>startAngle</code> and extends
     * for <code>arcAngle</code> degrees.
     * Angles are interpreted such that 0&nbsp;degrees
     * is at the 3&nbsp;o'clock position.
     * A positive value indicates a counter-clockwise rotation
     * while a negative value indicates a clockwise rotation.
     * <p>
     * The center of the arc is the center of the rectangle whose origin
     * is (<i>x</i>,&nbsp;<i>y</i>) and whose size is specified by the
     * <code>width</code> and <code>height</code> arguments.
     * <p>
     * The angles are specified relative to the non-square extents of
     * the bounding rectangle such that 45 degrees always falls on the
     * line from the center of the ellipse to the upper right corner of
     * the bounding rectangle. As a result, if the bounding rectangle is
     * noticeably longer in one axis than the other, the angles to the
     * start and end of the arc segment will be skewed farther along the
     * longer axis of the bounds.
     * @param        x the <i>x</i> coordinate of the
     *                    upper-left corner of the arc to be filled.
     * @param        y the <i>y</i>  coordinate of the
     *                    upper-left corner of the arc to be filled.
     * @param        width the width of the arc to be filled.
     * @param        height the height of the arc to be filled.
     * @param        startAngle the beginning angle.
     * @param        arcAngle the angular extent of the arc,
     *                    relative to the start angle.
     */
    public void fillArc(double x, double y, double width, double height, double startAngle, double arcAngle)
    {
        this.drawFillArc(x, y, width, height, startAngle, arcAngle, true);
    }
    
    
    
    /**
     * Draws or fills a circular or elliptical arc
     * covering the specified rectangle.
     * <p>
     * The resulting arc begins at <code>startAngle</code> and extends
     * for <code>arcAngle</code> degrees, using the current color.
     * Angles are interpreted such that 0&nbsp;degrees
     * is at the 3&nbsp;o'clock position.
     * A positive value indicates a counter-clockwise rotation
     * while a negative value indicates a clockwise rotation.
     * <p>
     * The center of the arc is the center of the rectangle whose origin
     * is (<i>x</i>,&nbsp;<i>y</i>) and whose size is specified by the
     * <code>width</code> and <code>height</code> arguments.
     * <p>
     * The angles are specified relative to the non-square extents of
     * the bounding rectangle such that 45 degrees always falls on the
     * line from the center of the ellipse to the upper right corner of
     * the bounding rectangle. As a result, if the bounding rectangle is
     * noticeably longer in one axis than the other, the angles to the
     * start and end of the arc segment will be skewed farther along the
     * longer axis of the bounds.
     * @param        x the <i>x</i> coordinate of the
     *                    upper-left corner of the arc to be drawn.
     * @param        y the <i>y</i>  coordinate of the
     *                    upper-left corner of the arc to be drawn.
     * @param        width the width of the arc to be drawn.
     * @param        height the height of the arc to be drawn.
     * @param        startAngle the beginning angle, in degrees
     * @param        sweepAngle the angular extent of the arc, in degrees,
     *                    relative to the start angle.
     *  @param       filled  True if arc to be filled as a wedge, false if just arc outline                  
     */
    private void drawFillArc(double x, double y, double width, double height, double startAngle, double sweepAngle, boolean filled)
    {
        ODGGraphicsStyle style;
        if (filled)
            style = odgDocument.getGraphicsStyle(this, false, true);
        else
            style = odgDocument.getGraphicsStyle(this, true, false);
        
        
        if ((!ODGEntity.preApplyTransform) || (javaTransform.isIdentity()))
        {
            // either not pre-applying the transform, or it's the identity - life is simple
            if (width == height)
            {
                // circular arc
                double radius = width/2;
                RealPoint center = new RealPoint(x + radius, y + radius, 0);
                ODGCircle odgCircle = new ODGCircle(center, radius, startAngle, sweepAngle, style, javaTransform);
                odgDocument.addEntity(odgCircle);
            }
            else
            {
                // elliptical arc
                double xRadius = width/2;
                double yRadius = height/2;
                RealPoint center = new RealPoint(x + xRadius, y + yRadius, 0);
                ODGEllipse odgEllipse = new ODGEllipse(center, xRadius, yRadius, startAngle, sweepAngle, style, javaTransform);
                odgDocument.addEntity(odgEllipse);
            }
        }
        else
        {
            // Hoo boy - pre-applying transform, and it's not the identity - circles can become ellipses,
            // and ellipses can be rotated by weird amounts, especially when shear is involved.
            // And ODGEllipse is a shape that is oriented with its major and minor axes horizontal and vertical,
            // expecting any rotation to be supplied in the transform (which fortunately DOES work correctly in LibreOffice)
            // So we have some unpleasant work to do... luckily I have my feathers numbered for just such an occasion...
            ODGEntity odgArc = createTransformedODGEllipticalArc(x, y, width, height, startAngle, sweepAngle, style);
            odgDocument.addEntity(odgArc);
        }
        
    }
    
    
    
    private ODGEntity createTransformedODGEllipticalArc(double x, double y, double width, double height, double startAngle, double sweepAngle, ODGGraphicsStyle style)
    {
        
        // BTW, it's worth noting that this math uses the standard convention where a positive angle (parameter) moves the positive
        // x-axis toward the positive y-axis. With both Java and ODG coordinates, this rotates clockwise. This is how Java
        // and ODG transformations work. However, note that the start and sweep angles of an arc are specified as counter-clockwise 
        // rotations. Thus we negate the angle info coming in, and re-negate going out.
        
        // get oval parameters
        double xCen = x + width/2;
        double yCen = y + height/2;
        
        double a = width/2;
        double b = height/2;
        
        double startParameter = -(startAngle * (Math.PI / 180));
        
        RealPoint center = new RealPoint(xCen, yCen, 0);
        RealPoint startPoint = new RealPoint(xCen + a*Math.cos(startParameter), yCen + b*Math.sin(startParameter), 0);
        
        
        // Here goes the complex stuff for the general transformation case involving rotation and skewing of the ellipse    
       
        // Finding the new major axis under a linear transformation is actually a bit tricky: though an
        // ellipse is mapped to an ellipse under a linear (even affine) transformation, the major axis 
        // of the original ellipse is not mapped to the transformed ellipse's major axis by the transformation.
        // To find the new major axis, we do the following: represent points on the ellipse parametrically, as
        //   (x, y) = (a*cos(t), b*sin(t)) (which we'll use as a column vector). 
        // Representing the linear transformation by the matrix 
        //   (c11, c12)
        //   (c21, c22), 
        // the points of the linearly transformed ellipse become
        //   (x, y) = (c11*a*cos(t) + c12*b*sin(t), c21*a*cos(t) + c22*b*sin(t)).
        // Now the points on the major axis are the points farthest from the center (here 
        // at (0, 0) since we're using just the linear part of the affine transfomation). We can
        // then find these point by solving for t that maximizes the (square of) the distance from the center:
        //   D  =  x^2 + y^2  =  (c11*a*cos(t) + c12*b*sin(t))^2  +  (c21*a*cos(t) + c22*b*sin(t))^2
        // i.e., solve for t such that dD/dt = 0.
        //   dD/dt  =  2 * (c11*a*cos(t) + c12*b*sin(t)) * (-c11*a*sin(t) + c12*b*cos(t))
        //           + 2 * (c21*a*cos(t) + c22*b*sin(t)) * (-c21*a*sin(t) + c22*b*cos(t))
        //          =  2 * { [b^2*(c12^2 + c22^2) - a^2*(c11^2 + c21^2)] * sin(t) * cos(t)
        //                      + a*b*(c11*c12 + c21*c22) * (cos(t)^2 - sin(t)^2) }
        // Setting this equal to 0, and dividing by -2 and cos(t), gives
        //   A*tan(t)^2 + B*tan(t) - A  =  0,
        //      where  A  =  a*b*(c11*c12 + c21*c22)
        //      and    B  =  a^2*(c11^2 + c21^2) - b^2*(c12^2 + c22^2)
        // The quadratic formula then gives us our solution:
        //   tan(t)  =  (-B +/- sqrt(B^2 - 4*A*C))/(2*A)  =  (-B +/- sqrt(B^2 + 4*A^2))/(2*A)
        //
        
        
        // get matrix coefficients
        double[] basisVects = {1, 0, 0, 1};
        mapVectors(javaTransform, basisVects);
        double c11 = basisVects[0];
        double c12 = basisVects[2];
        double c21 = basisVects[1];
        double c22 = basisVects[3];
        
        double A = a*b*(c11*c12 + c21*c22);
        double B = a*a*(c11*c11 + c21*c21) - b*b*(c12*c12 + c22*c22);
        
        double tangent1;
        if (A != 0)
        {
            tangent1 = (-B + Math.sqrt(B*B + 4*A*A))/(2*A);
        }
        else
        {
            tangent1 = 0;
        }
        
        double t1 = Math.atan(tangent1);
        double t2 = t1 + Math.PI/2;
        
        double majorAxisParameter;
        double minorAxisParameter;
        double majorAxisLength;
        double minorAxisLength;
        
        // see which is the major axis
        double D1 = ((c11*a*Math.cos(t1) + c12*b*Math.sin(t1))*(c11*a*Math.cos(t1) + c12*b*Math.sin(t1))  +  (c21*a*Math.cos(t1) + c22*b*Math.sin(t1))*(c21*a*Math.cos(t1) + c22*b*Math.sin(t1)));
        double D2 = ((c11*a*Math.cos(t2) + c12*b*Math.sin(t2))*(c11*a*Math.cos(t2) + c12*b*Math.sin(t2))  +  (c21*a*Math.cos(t2) + c22*b*Math.sin(t2))*(c21*a*Math.cos(t2) + c22*b*Math.sin(t2)));
        
        if (D1 >= D2)
        {
            majorAxisParameter = t1;
            minorAxisParameter = t1 + Math.PI/2;
            majorAxisLength = Math.sqrt(D1);
            minorAxisLength = Math.sqrt(D2);
        }
        else
        {
            majorAxisParameter = t2;
            minorAxisParameter = t2 + Math.PI/2;
            majorAxisLength = Math.sqrt(D2);
            minorAxisLength = Math.sqrt(D1);
        }
        
        double transformedAxisRatio = minorAxisLength/majorAxisLength;
        
        
        // get the transformed center, major and (relative) minor aces
        RealPoint transformedCenter = RealPoint.mapPoint(javaTransform, center);
        RealPoint transformedMajorAxisRelative = new RealPoint(c11*a*Math.cos(majorAxisParameter) + c12*b*Math.sin(majorAxisParameter), c21*a*Math.cos(majorAxisParameter) + c22*b*Math.sin(majorAxisParameter), 0);
        RealPoint transformedMinorAxisRelative = new RealPoint(c11*a*Math.cos(minorAxisParameter) + c12*b*Math.sin(minorAxisParameter), c21*a*Math.cos(minorAxisParameter) + c22*b*Math.sin(minorAxisParameter), 0);
        
        // get transformed start and end points for center lines if needed, as well as to find corresponding ellipse parameters
        RealPoint transformedStartPoint = RealPoint.mapPoint(javaTransform, startPoint);
        RealPoint transformedStartPointRelative = RealPoint.difference(transformedStartPoint, transformedCenter);
        
        // want the arc that goes between the start and end points; but need to supply the start and end parametric t values,
        // and these are relative to the major axis. So we need to find the t values corresponding to the start and end points.
        // To do this, we decompose relative to the major and minor axes and use the basic ellipse equation to find the corresponding t values.
        // Let (u, v) represent the coordinates of a point on the ellipse in the major/minor axis coordinate system. Then
        //    (u, v)  =  (majorAxisLength*cos(t), minorAxisLength*sin(t)),  or
        //    v/u  =  minorAxisLength/majorAxisLength*tan(t), or   t = atan((v*majorAxisLength)/(u*minorAxisLength)), where
        //    u  =  startVector dotproduct majorAxisVector/magnitude(majorAxisVector), and v  =  same with minorAxisVector
        
        double transformedStartParameter;
        
        double uStart = RealPoint.dotProduct(transformedStartPointRelative, transformedMajorAxisRelative)/RealPoint.magnitude(transformedMajorAxisRelative);
        double vStart = RealPoint.dotProduct(transformedStartPointRelative, transformedMinorAxisRelative)/RealPoint.magnitude(transformedMinorAxisRelative);
        if (uStart == 0)
        {
            if (vStart > 0)
                transformedStartParameter = Math.PI/2;
            else
                transformedStartParameter = -Math.PI/2;
        }
        else
        {
            transformedStartParameter = Math.atan((vStart*majorAxisLength)/(uStart*minorAxisLength));
            if (uStart < 0)
            {
                transformedStartParameter += Math.PI;
            }
        }
        
        // finally, create an ellipse or arc or circle with the given parameters
        if (transformedAxisRatio == 1)
        {
            // we have a circle or circular arc
            double radius = RealPoint.magnitude(transformedMajorAxisRelative);
            
            if (Math.abs(sweepAngle) >= 360)
            {
                // full circle
                return new ODGCircle(transformedCenter, radius, 0, 360, style, new AffineTransform());
            }
            else
            {
                
                // circular arc - need to determine start and end parameters relative to standard axes
                double majorAxisAngle = RealPoint.angleBetween(new RealPoint(1,0,0), transformedMajorAxisRelative);
                if (transformedMajorAxisRelative.y < 0)
                    majorAxisAngle = -majorAxisAngle;
                double arcStartAngle = (majorAxisAngle + transformedStartParameter) * 180/Math.PI;
                return new ODGCircle(transformedCenter, radius, -arcStartAngle, sweepAngle, style, new AffineTransform());
                
            }
        }
        else
        {
            double xRadius = RealPoint.magnitude(transformedMajorAxisRelative);
            double yRadius = RealPoint.magnitude(transformedMinorAxisRelative);
            double arcStartAngle = (transformedStartParameter) * 180/Math.PI;
            
            // need to add a rotation transform, with a translation, and set center to 0
            double majorAxisAngle = RealPoint.angleBetween(new RealPoint(1,0,0), transformedMajorAxisRelative);
            if (transformedMajorAxisRelative.y < 0)
                majorAxisAngle = -majorAxisAngle;
            AffineTransform transform = new AffineTransform();
            transform.translate(transformedCenter.x, transformedCenter.y);
            transform.rotate(majorAxisAngle);
            
            return new ODGEllipse(new RealPoint(0,0,0), xRadius, yRadius, -arcStartAngle, sweepAngle, style, transform);
        }
    }

    /**
     * Draws a sequence of connected lines defined by
     * arrays of <i>x</i> and <i>y</i> coordinates.
     * Each pair of (<i>x</i>,&nbsp;<i>y</i>) coordinates defines a point.
     * The figure is not closed if the first point
     * differs from the last point.
     * @param       xPoints an array of <i>x</i> points
     * @param       yPoints an array of <i>y</i> points
     * @param       nPoints the total number of points
     */
    public void drawPolyline(int xPoints[], int yPoints[], int nPoints)
    {
        double[] xPointsFloat = new double[nPoints];
        double[] yPointsFloat = new double[nPoints];
        for (int i = 0; i < nPoints; i++)
        {
            xPointsFloat[i] = xPoints[i];
            yPointsFloat[i] = yPoints[i];
        }
        
        this.drawPolyline(xPointsFloat, yPointsFloat, nPoints);
    }
    
    
    /**
     * Draws a sequence of connected lines defined by
     * arrays of <i>x</i> and <i>y</i> coordinates.
     * Each pair of (<i>x</i>,&nbsp;<i>y</i>) coordinates defines a point.
     * The figure is not closed if the first point
     * differs from the last point.
     * @param       xPoints an array of <i>x</i> points
     * @param       yPoints an array of <i>y</i> points
     * @param       nPoints the total number of points
     */
    public void drawPolyline(double xPoints[], double yPoints[], int nPoints)
    {
        // close if last point equals first point
        boolean isClosed;
        if((xPoints[0] == xPoints[nPoints-1]) && (yPoints[0] == yPoints[nPoints-1]))
        {
            isClosed = true;
        }
        else
        {
            isClosed = false;
        }
        
        drawPolyline(xPoints, yPoints, nPoints, isClosed);
    }
    
    
    /**
     * Draws a sequence of connected lines defined by
     * arrays of <i>x</i> and <i>y</i> coordinates.
     * Each pair of (<i>x</i>,&nbsp;<i>y</i>) coordinates defines a point.
     * The isClosed parameter indicates if the polygon should be closed,
     * i.e., have a line segment added joining the first and last points
     * if they are different.
     * @param       xPoints an array of <i>x</i> points
     * @param       yPoints an array of <i>y</i> points
     * @param       nPoints the total number of points
     * @param       isClosed - indicate if the polygon should be closed
     */
    private void drawPolyline(double xPoints[], double yPoints[], int nPoints, boolean isClosed)
    {
        drawFillPolyline(xPoints, yPoints, nPoints, isClosed, false /*isFilled*/);
    }
    
    
    /**
     * Draws a sequence of connected lines defined by
     * arrays of <i>x</i> and <i>y</i> coordinates.
     * Each pair of (<i>x</i>,&nbsp;<i>y</i>) coordinates defines a point.
     * The isClosed parameter indicates if the polygon should be closed,
     * i.e., have a line segment added joining the first and last points
     * if they are different.
     * @param       xPoints an array of <i>x</i> points
     * @param       yPoints an array of <i>y</i> points
     * @param       nPoints the total number of points
     * @param       isClosed - indicate if the polygon should be closed
     */
    private void drawFillPolyline(double xPoints[], double yPoints[], int nPoints, boolean isClosed, boolean isFilled)
    {
        ODGGraphicsStyle style;
        if (isFilled)
            style = odgDocument.getGraphicsStyle(this, false, true);
        else
            style = odgDocument.getGraphicsStyle(this, true, false);
        
        // create array of points for transforming if needed
        double[] points = new double[2*nPoints];
        for (int i = 0; i < nPoints; i++)
        {
            points[2*i] = xPoints[i];
            points[2*i+1] = yPoints[i];
        }
        
        // see if we're pre-applying the transform to work around LibreOffice transform bugs
        AffineTransform transform;
        if (ODGEntity.preApplyTransform)
        {
            // apply Java transform to points, and use identity transform with resulting shape
            javaTransform.transform(points, 0, points, 0, nPoints);
            transform = new AffineTransform();
        }
        else
        {
            transform = javaTransform;
        }
        
        Vector<RealPoint> vertices = new Vector<RealPoint>();
        for (int i = 0; i < nPoints; i++)
        {
            vertices.add(new RealPoint(points[2*i], points[2*i+1], 0));
        }
        
        if (isClosed)
        {
            // represent as a polygon
            ODGPolygon polygon = new ODGPolygon(vertices, style, transform);
            odgDocument.addEntity(polygon);
        }
        else
        {
            // general polyline
            ODGPolyline polyline = new ODGPolyline(vertices, style, transform);
            odgDocument.addEntity(polyline);
        }
        
        
    }

    /**
     * Draws a closed polygon defined by
     * arrays of <i>x</i> and <i>y</i> coordinates.
     * Each pair of (<i>x</i>,&nbsp;<i>y</i>) coordinates defines a point.
     * <p>
     * This method draws the polygon defined by <code>nPoint</code> line
     * segments, where the first <code>nPoint&nbsp;-&nbsp;1</code>
     * line segments are line segments from
     * <code>(xPoints[i&nbsp;-&nbsp;1],&nbsp;yPoints[i&nbsp;-&nbsp;1])</code>
     * to <code>(xPoints[i],&nbsp;yPoints[i])</code>, for
     * 1&nbsp;&le;&nbsp;<i>i</i>&nbsp;&le;&nbsp;<code>nPoints</code>.
     * The figure is automatically closed by drawing a line connecting
     * the final point to the first point, if those points are different.
     * @param        xPoints   a an array of <code>x</code> coordinates.
     * @param        yPoints   a an array of <code>y</code> coordinates.
     * @param        nPoints   a the total number of points.
     */
    public void drawPolygon(int xPoints[], int yPoints[], int nPoints)
    {
        double[] xPointsFloat = new double[nPoints];
        double[] yPointsFloat = new double[nPoints];
        for (int i = 0; i < nPoints; i++)
        {
            xPointsFloat[i] = xPoints[i];
            yPointsFloat[i] = yPoints[i];
        }
        
        this.drawPolygon(xPointsFloat, yPointsFloat, nPoints);
    }
    
    
    /**
     * Draws a closed polygon defined by
     * arrays of <i>x</i> and <i>y</i> coordinates.
     * Each pair of (<i>x</i>,&nbsp;<i>y</i>) coordinates defines a point.
     * <p>
     * This method draws the polygon defined by <code>nPoint</code> line
     * segments, where the first <code>nPoint&nbsp;-&nbsp;1</code>
     * line segments are line segments from
     * <code>(xPoints[i&nbsp;-&nbsp;1],&nbsp;yPoints[i&nbsp;-&nbsp;1])</code>
     * to <code>(xPoints[i],&nbsp;yPoints[i])</code>, for
     * 1&nbsp;&le;&nbsp;<i>i</i>&nbsp;&le;&nbsp;<code>nPoints</code>.
     * The figure is automatically closed by drawing a line connecting
     * the final point to the first point, if those points are different.
     * @param        xPoints   a an array of <code>x</code> coordinates.
     * @param        yPoints   a an array of <code>y</code> coordinates.
     * @param        nPoints   a the total number of points.
     */
    public void drawPolygon(double xPoints[], double yPoints[], int nPoints)
    {
        this.drawPolyline(xPoints, yPoints, nPoints, true);
    }

    
    /**
     * Draws the outline of a polygon defined by the specified
     * <code>Polygon</code> object.
     * @param        p the polygon to draw.
     */
    public void drawPolygon(Polygon p) 
    {
        drawPolygon(p.xpoints, p.ypoints, p.npoints);
    }

    
    /**
     * Fills a closed polygon defined by
     * arrays of <i>x</i> and <i>y</i> coordinates.
     * <p>
     * This method fills the polygon defined by <code>nPoint</code> line
     * segments, where the first <code>nPoint&nbsp;-&nbsp;1</code>
     * line segments are line segments from
     * <code>(xPoints[i&nbsp;-&nbsp;1],&nbsp;yPoints[i&nbsp;-&nbsp;1])</code>
     * to <code>(xPoints[i],&nbsp;yPoints[i])</code>, for
     * 1&nbsp;&le;&nbsp;<i>i</i>&nbsp;&le;&nbsp;<code>nPoints</code>.
     * The figure is automatically closed by drawing a line connecting
     * the final point to the first point, if those points are different.
     * @param        xPoints   a an array of <code>x</code> coordinates.
     * @param        yPoints   a an array of <code>y</code> coordinates.
     * @param        nPoints   a the total number of points.
     */
    public void fillPolygon(int xPoints[], int yPoints[], int nPoints)
    {
        double[] xPointsDouble = new double[nPoints];
        double[] yPointsDouble = new double[nPoints];
        for (int i = 0; i < nPoints; i++)
        {
            xPointsDouble[i] = xPoints[i];
            yPointsDouble[i] = yPoints[i];
        }
        
        this.fillPolygon(xPointsDouble, yPointsDouble, nPoints);
        
    }
    
    
    /**
     * Fills a closed polygon defined by
     * arrays of <i>x</i> and <i>y</i> coordinates.
     * <p>
     * This method fills the polygon defined by <code>nPoint</code> line
     * segments, where the first <code>nPoint&nbsp;-&nbsp;1</code>
     * line segments are line segments from
     * <code>(xPoints[i&nbsp;-&nbsp;1],&nbsp;yPoints[i&nbsp;-&nbsp;1])</code>
     * to <code>(xPoints[i],&nbsp;yPoints[i])</code>, for
     * 1&nbsp;&le;&nbsp;<i>i</i>&nbsp;&le;&nbsp;<code>nPoints</code>.
     * The figure is automatically closed by drawing a line connecting
     * the final point to the first point, if those points are different.
     * @param        xPoints   a an array of <code>x</code> coordinates.
     * @param        yPoints   a an array of <code>y</code> coordinates.
     * @param        nPoints   a the total number of points.
     */
    public void fillPolygon(double xPoints[], double yPoints[], int nPoints)
    {
        drawFillPolyline(xPoints, yPoints, nPoints, true /*isClosed*/, true /*isFilled*/);
    }

    
    /**
     * Fills the polygon defined by the specified Polygon object with
     * the graphics context's current color.
     * @param        p the polygon to fill.
     */
    public void fillPolygon(Polygon p) {
        fill(p);
    }
    
    
    
    /**
     * An extension method that supports the drawing of a B-spline. For the common case of a spline that passes through 
     * the first and last control points, set the multiplicity of these equal to the spline degree plus 1.
     * 
     * @param degree        The degree of the spline; must be greater than 0
     * @param controlPoints The control points that define the shape of the spline, {x1, y1, x2, y2, ..., xn, yn}.
     * @param multiplicities    The multiplicities for the control points; the number of multiplicity values must equal the
     *                      number of control points (i.e., the array size must be half that of the controlPoints array)
     * @param throughEndpoints  If true, the spline will be forced to pass through the endpoints by setting the end
     *                      control point multiplicities to degree + 1
     */
    /*
    public void drawSpline(int degree, double[] controlPoints, int[] multiplicities, boolean throughEndpoints)
    {
        Vector<SplineControlPoint> transformedControlPoints = new Vector<SplineControlPoint>();
        
        // transform the control points according to the current transform matrix
        for (int i = 0; i < controlPoints.length/2; i++)
        {
            double[] coords = new double[]{controlPoints[2*i], controlPoints[2*i+1]};
            graphicsMatrix.transform(coords, 0, coords, 0, 1);
            SplineControlPoint transformedControlPoint = new SplineControlPoint(coords[0], coords[1], 0, multiplicities[i]);
            transformedControlPoints.add(transformedControlPoint);
        }
        
        ODGSpline spline = new ODGSpline(degree, transformedControlPoints, throughEndpoints, this);
        odgDocument.addEntity(spline);
    }
    */
    
    
    
    /**
     * An extension method that supports the drawing of a B-spline with an arbitrary knot sequence. This is included
     * specifically to cover the case of the quadratic and cubic curve segments generated by PathIterator, which are
     * specified as Bezier curves, or B-splines with knots {0,0,0,0,1,1,1,1} (for the cubic). This will handle other
     * knot sequences if needed; however, note the constraint on the number of knots vs the number of control points.
     * Also, the multiplicities of the control points are all set to 1.
     * 
     * @param degree        The degree of the spline; must be greater than 0
     * @param controlPoints The control points that define the shape of the spline, {x1, y1, x2, y2, ..., xn, yn}.
     * @param knots         The knots for the control points; the number of knot values must equal the
     *                      number of control points plus the degree plus 1. Setting the first degree + 1 knots equal 
     *                      and the last degree + 1 knots equal will make the spline pass through the endpoints
     */
    /*
    public void drawSpline(int degree, double[] controlPoints, double[] knots)
    {
        // transform the control points according to the current transform matrix
        graphicsMatrix.transform(controlPoints, 0, controlPoints, 0, controlPoints.length/2);
        ODGSpline spline = new ODGSpline(degree, controlPoints, knots, this);
        odgDocument.addEntity(spline);
    }
    */
    

    /**
     * Draws the text given by the specified string, using this
     * graphics context's current font and color. The baseline of the
     * leftmost character is at position (<i>x</i>,&nbsp;<i>y</i>) in this
     * graphics context's coordinate system.
     * @param       str      the string to be drawn.
     * @param       x        the <i>x</i> coordinate.
     * @param       y        the <i>y</i> coordinate.
     * @throws NullPointerException if <code>str</code> is <code>null</code>.
     */
    public void drawString(String str, int x, int y)
    {
        drawString(str, (double)x, (double)y);
    }

    /**
     * Renders the text of the specified iterator, but ignoring text attributes.
     * <p>
     * The baseline of the leftmost character is at position
     * (<i>x</i>,&nbsp;<i>y</i>) in this graphics context's coordinate system.
     * @param       charIterator   the iterator whose text is to be drawn
     * @param       x        the <i>x</i> coordinate.
     * @param       y        the <i>y</i> coordinate.
     * @throws NullPointerException if <code>iterator</code> is
     * <code>null</code>.
     */
   public void drawString(AttributedCharacterIterator charIterator, int x, int y)
   {
       if (charIterator == null)
       {
           throw new NullPointerException();
       }
       else
       {
           // Just use String method
           StringBuffer stringBuffer = new StringBuffer();
           
           for(char c = charIterator.first(); c != CharacterIterator.DONE; c = charIterator.next()) 
           {
               stringBuffer.append(c);
           }
           
           drawString(stringBuffer.toString(), x, y);
       }
   }

    /**
     * Draws the text given by the specified character array, using this
     * graphics context's current font and color. The baseline of the
     * first character is at position (<i>x</i>,&nbsp;<i>y</i>) in this
     * graphics context's coordinate system.
     * @param data the array of characters to be drawn
     * @param offset the start offset in the data
     * @param length the number of characters to be drawn
     * @param x the <i>x</i> coordinate of the baseline of the text
     * @param y the <i>y</i> coordinate of the baseline of the text
     * @throws NullPointerException if <code>data</code> is <code>null</code>.
     * @throws IndexOutOfBoundsException if <code>offset</code> or
     * <code>length</code>is less than zero, or
     * <code>offset+length</code> is greater than the length of the
     * <code>data</code> array.
     * @see         java.awt.Graphics#drawBytes
     * @see         java.awt.Graphics#drawString
     */
    public void drawChars(char data[], int offset, int length, int x, int y) 
    {
        drawString(new String(data, offset, length), x, y);
    }

    
    /**
     * Draws the text given by the specified byte array, using this
     * graphics context's current font and color. The baseline of the
     * first character is at position (<i>x</i>,&nbsp;<i>y</i>) in this
     * graphics context's coordinate system.
     * <p>
     * Use of this method is not recommended as each byte is interpreted
     * as a Unicode code point in the range 0 to 255, and so can only be
     * used to draw Latin characters in that range.
     * @param data the data to be drawn
     * @param offset the start offset in the data
     * @param length the number of bytes that are drawn
     * @param x the <i>x</i> coordinate of the baseline of the text
     * @param y the <i>y</i> coordinate of the baseline of the text
     * @throws NullPointerException if <code>data</code> is <code>null</code>.
     * @throws IndexOutOfBoundsException if <code>offset</code> or
     * <code>length</code>is less than zero, or <code>offset+length</code>
     * is greater than the length of the <code>data</code> array.
     */
    @SuppressWarnings("deprecation")
    public void drawBytes(byte data[], int offset, int length, int x, int y) {
        drawString(new String(data, 0, offset, length), x, y);
    }

    
    /**
     * Throws UnsupportedOperationException, as drawImage not supported.
     */
    public boolean drawImage(Image img, int x, int y, ImageObserver observer)
    {
        // Not supported
        throw new UnsupportedOperationException("Image drawing not supported");
    }

    
    /**
     * Throws UnsupportedOperationException, as drawImage not supported.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height, ImageObserver observer)
    {
        // Not supported
        throw new UnsupportedOperationException("Image drawing not supported");
    }
    
    
    
    /**
     * Throws UnsupportedOperationException, as drawImage not supported.
     */
    public boolean drawImage(Image img, int x, int y,
                                      Color bgcolor,
                                      ImageObserver observer)
    {
        // Not supported
        throw new UnsupportedOperationException("Image drawing not supported");
    }

    
    /**
     * Throws UnsupportedOperationException, as drawImage not supported.
     */
    public boolean drawImage(Image img, int x, int y,
                                      int width, int height,
                                      Color bgcolor,
                                      ImageObserver observer)
    {
        // Not supported
        throw new UnsupportedOperationException("Image drawing not supported");
    }

    
    /**
     * Throws UnsupportedOperationException, as drawImage not supported.
     */
    public boolean drawImage(Image img,
                                      int dx1, int dy1, int dx2, int dy2,
                                      int sx1, int sy1, int sx2, int sy2,
                                      ImageObserver observer)
    {
        // Not supported
        throw new UnsupportedOperationException("Image drawing not supported");
    }
    
    
    
    /**
     * Throws UnsupportedOperationException, as drawImage not supported.
     */
    public boolean drawImage(Image img,
                                      int dx1, int dy1, int dx2, int dy2,
                                      int sx1, int sy1, int sx2, int sy2,
                                      Color bgcolor,
                                      ImageObserver observer)
    {
        // Not supported
        throw new UnsupportedOperationException("Image drawing not supported");
    }
    
    
    
    
    /**
     * Disposes of this graphics context and releases
     * any system resources that it is using. Does nothing here.
     */
    public void dispose()
    {
        // nothing special here to do...
    }

    
    /**
     * Disposes of this graphics context once it is no longer referenced.
     * @see #dispose
     */
    public void finalize() 
    {
        dispose();
    }

    
    /**
     * Returns a <code>String</code> object representing this
     *                        <code>Graphics</code> object's value.
     * @return       a string representation of this graphics context.
     */
    public String toString() 
    {
        return getClass().getName() + "[font=" + getFont() + ",color=" + getColor() + "]";
    }

    
    /**
     * Returns null as clipping not supported.
     * @return      null
     */
    @Deprecated
    public Rectangle getClipRect() {
        return getClipBounds();
    }

    
    /**
     * Returns true as clipping not supported.
     * @return true.
     */
    public boolean hitClip(int x, int y, int width, int height) 
    {    
        return true;
    }

    
    /**
     * Returns r, as clipping not supported.
     * @param  r    the rectangle where the current clipping area is
     *              copied to.  Any current values in this rectangle are
     *              overwritten.
     * @return      r.
     */
    public Rectangle getClipBounds(Rectangle r) 
    {
        return r;
    }
    
    
    /**
     * Does nothing...
     */
    @Override
    public void addRenderingHints(Map<?, ?> arg0)
    {
        // do nothing
        
    }

    /**
     * Does nothing as clipping not supported.
     */
    @Override
    public void clip(Shape arg0)
    {
        // do nothing no clipping
        
    }

    
    /**
     * Draw the specified shape.
     * 
     * @param   shape  shape to be drawn
     */
    @Override
    public void draw(Shape shape)
    {
        
        // branch based on the type of shape
        if (shape instanceof Line2D)
        {
            Line2D line = (Line2D)shape;
            drawLine(line.getX1(), line.getY1(), line.getX2(), line.getY2());
        }
        else if (shape instanceof Ellipse2D)
        {
            Ellipse2D ellipse = (Ellipse2D)shape;
            Rectangle2D rect = ellipse.getFrame();
            this.drawOval(rect.getX(), rect.getY(), rect.getWidth(), rect.getHeight());
        }
        else if (shape instanceof Arc2D)
        {
            Arc2D arc = (Arc2D)shape;
            Rectangle2D rect = arc.getFrame();
            this.drawArc(rect.getX(), rect.getY(), rect.getWidth(), rect.getHeight(), arc.getAngleStart(), arc.getAngleExtent());
        }
        else if (shape instanceof Rectangle2D)
        {
            Rectangle2D rect = (Rectangle2D)shape;
            this.drawRect(rect.getX(), rect.getY(), rect.getWidth(), rect.getHeight());
        }
        else if (shape instanceof RoundRectangle2D)
        {
            RoundRectangle2D rect = (RoundRectangle2D)shape;
            this.drawRoundRect(rect.getX(), rect.getY(), rect.getWidth(), rect.getHeight(), rect.getArcWidth(), rect.getArcHeight());
        }
        else
        {
            // general shape - create as general ODG path
            ODGGraphicsStyle style = odgDocument.getGraphicsStyle(this, true, false);
            ODGPath odgPath = new ODGPath(shape, style, javaTransform);
            odgDocument.addEntity(odgPath);
        }   
        
    }
    
    
    

    
    /**
     * Throws UnsupportedOperationException as glyph drawing not supported.
     */
    @Override
    public void drawGlyphVector(GlyphVector arg0, float arg1, float arg2)
    {
        throw new UnsupportedOperationException("Glyph drawing not supported");
    }

    
    /**
     * Throws UnsupportedOperationException as image drawing not supported.
     */
    @Override
    public boolean drawImage(Image arg0, AffineTransform arg1, ImageObserver arg2)
    {
        // Not supported
        throw new UnsupportedOperationException("Image drawing not supported");
    }

    
    /**
     * Throws UnsupportedOperationException as image drawing not supported.
     */
    @Override
    public void drawImage(BufferedImage arg0, BufferedImageOp arg1, int arg2, int arg3)
    {
        // Not supported
        throw new UnsupportedOperationException("Image drawing not supported");
    }

    
    /**
     * Throws UnsupportedOperationException as image drawing not supported.
     */
    @Override
    public void drawRenderableImage(RenderableImage arg0, AffineTransform arg1)
    {
        // Not supported
        throw new UnsupportedOperationException("Image drawing not supported");
    }

    
    /**
     * Throws UnsupportedOperationException as image drawing not supported.
     */
    @Override
    public void drawRenderedImage(RenderedImage arg0, AffineTransform arg1)
    {
        // Not supported
        throw new UnsupportedOperationException("Image drawing not supported");
    }

    @Override
    public void drawString(String str, float x, float y)
    {
        drawString(str, (double)x, (double)y);
    }
    
    
    /**
     * Draws the text given by the specified string, using this
     * graphics context's current font and color. The baseline of the
     * leftmost character is at position (<i>x</i>,&nbsp;<i>y</i>) in this
     * graphics context's coordinate system.
     * @param       string   the string to be drawn.
     * @param       x        the <i>x</i> coordinate.
     * @param       y        the <i>y</i> coordinate.
     * @throws NullPointerException if <code>str</code> is <code>null</code>.
     */
    public void drawString(String string, double x, double y)
    {
        if (string == null)
        {
            throw new NullPointerException();
        }
        else
        {
            // for text, need a box that is neither stroked nor filled
            ODGGraphicsStyle style = odgDocument.getGraphicsStyle(this, false, false);
            ODGTextStyle textStyle = odgDocument.getTextStyle(this);
            ODGFrame text = new ODGFrame(string, new RealPoint(x, y, 0), style, textStyle, this, javaTransform);
            odgDocument.addEntity(text);
        }
    }
    
    

    @Override
    public void drawString(AttributedCharacterIterator charIterator, float x, float y)
    {
        if (charIterator == null)
        {
            throw new NullPointerException();
        }
        else
        {
            // Just use String method
            StringBuffer stringBuffer = new StringBuffer();
            
            for(char c = charIterator.first(); c != CharacterIterator.DONE; c = charIterator.next()) 
            {
                stringBuffer.append(c);
            }
            
            drawString(stringBuffer.toString(), x, y);
        }
    }

    
    /**
     * Fill the specified shape with the current foreground color.
     * 
     * @param   shape  shape to be filled
     */
    @Override
    public void fill(Shape shape)
    {
        fill(shape, this.getColor());
    }
    
    
    /**
     * Clear the specified shape by filling with the current background color.
     * 
     * @param   shape  shape to be cleared
     */
    public void clear(Shape shape)
    {
        fill(shape, this.getBackground());
    }
    
    
    /**
     * Fill the specified shape with the specified color.
     * 
     * @param   shape  shape to be filled
     * @param   color  color to be used to fill shape
     */
    public void fill(Shape shape, Color color)
    {
        
        Color origColor = this.getColor();
        this.setColor(color);
        
        // branch based on the type of shape
        if (shape instanceof Line2D)
        {
            Line2D line = (Line2D)shape;
            drawLine(line.getX1(), line.getY1(), line.getX2(), line.getY2());
        }
        else if (shape instanceof Ellipse2D)
        {
            Ellipse2D ellipse = (Ellipse2D)shape;
            Rectangle2D rect = ellipse.getFrame();
            this.fillOval(rect.getX(), rect.getY(), rect.getWidth(), rect.getHeight());
        }
        else if (shape instanceof Arc2D)
        {
            Arc2D arc = (Arc2D)shape;
            Rectangle2D rect = arc.getFrame();
            this.fillArc(rect.getX(), rect.getY(), rect.getWidth(), rect.getHeight(), arc.getAngleStart(), arc.getAngleExtent());
        }
        else if (shape instanceof Rectangle2D)
        {
            Rectangle2D rect = (Rectangle2D)shape;
            this.fillRect(rect.getX(), rect.getY(), rect.getWidth(), rect.getHeight());
        }
        else
        {
            // render as general ODG path
            ODGGraphicsStyle style = odgDocument.getGraphicsStyle(this, false, true);
            ODGPath odgPath = new ODGPath(shape, style, javaTransform);
            odgDocument.addEntity(odgPath);
        }   
        
        this.setColor(origColor);
        
    }


    
    /**
     * Return the current background color.
     * 
     * @return  current background color
     */
    @Override
    public Color getBackground()
    {
        return backColor;
    }

    
    /**
     * Returns null; Composite not supported.
     */
    @Override
    public Composite getComposite()
    {
        // No composite
        return null;
    }

    
    /**
     * Returns null; GraphicsConfiguration not supported.
     */
    @Override
    public GraphicsConfiguration getDeviceConfiguration()
    {
        // Just return null
        return null;
    }

    
    /**
     * Returns null; Font render context needed for Font sizing; use from current graphics environment.
     */
    @Override
    public FontRenderContext getFontRenderContext()
    {
        // No font rendering context
        //return null;
        return GraphicsEnvironment.getLocalGraphicsEnvironment().createGraphics(new BufferedImage(100, 100, BufferedImage.TYPE_4BYTE_ABGR)).getFontRenderContext();
    }

    
    /**
     * Just return the current foreground color object.
     * 
     * @return  current foreground color
     */
    @Override
    public Paint getPaint()
    {
        // Return our color object
        return color;
    }

    
    /**
     * Return null; rendering hints not supported.
     */
    @Override
    public Object getRenderingHint(Key arg0)
    {
        // Return null
        return null;
    }

    
    /**
     * Return null; rendering hints not supported.
     */
    @Override
    public RenderingHints getRenderingHints()
    {
        // Just return null
        return null;
    }

    
    /**
     * Return our current pen stroke.
     * 
     * @return  current pen stroke
     */
    @Override
    public Stroke getStroke()
    {
        return stroke;
    }

    
    /**
     * Return a copy of the current graphics transformation being applied
     * to drawing and text operations.
     * 
     * @return  copy of current graphics transform
     */
    @Override
    public AffineTransform getTransform()
    {
        return new AffineTransform(javaTransform);
    }

    /**
     * Not supported; return false.
     */
    @Override
    public boolean hit(Rectangle arg0, Shape arg1, boolean arg2)
    {
        // Just return false
        return false;
    }

    
    /**
     * Apply the specified rotation to the current graphics transformation. Subsequent
     * graphics operations will have the new combined transformation applied.
     * 
     * @param   radians  rotation in radians
     */
    @Override
    public void rotate(double radians)
    {
        javaTransform.rotate(radians);
    }

    
    /**
     * Apply the specified rotation to the current graphics transformation, but first translating
     * by (x,y) and after translating back by (-x, -y). Subsequent graphics operations will have 
     * the new combined transformation applied.
     * 
     * @param   radians  rotation in radians
     * @param   x  x coordinate of pre- and post-translation
     * @param   y  y coordinate of pre- and post-translation
     */
    @Override
    public void rotate(double radians, double x, double y)
    {
        translate(x,y);
        rotate(radians);
        translate(-x,-y);
    }

    
    /**
     * Apply the specified scaling to the current graphics transformation. Subsequent
     * graphics operations will have the new combined transformation applied.
     * 
     * @param   sx  amount to scale x coordinates
     * @param   sy  amount to scale y coordinates
     */
    @Override
    public void scale(double sx, double sy)
    {
        javaTransform.scale(sx, sy);
    }

    
    /**
     * Set the background color used in clear operations to the specified color
     * 
     * @param   color  color to use for background
     */
    @Override
    public void setBackground(Color color)
    {
        backColor = color;
    }

    
    /**
     * Do nothing - composite not supported.
     */
    @Override
    public void setComposite(Composite arg0)
    {
        // Do nothing
    }

    
    /**
     * Use to set foreground color if instance of Color, otherwise do nothing.
     * 
     * @param   paint  paint object; only used if it's an instance of Color
     */
    @Override
    public void setPaint(Paint paint)
    {
        // Use to set Color if it's a color subclass, otherwise do nothing
        if (paint instanceof Color)
        {
            this.color = (Color)paint;
        }
        
    }

    
    /**
     * Do nothing; rendering hints not supported.
     */
    @Override
    public void setRenderingHint(Key arg0, Object arg1)
    {
        // Do nothing
    }

    
    /**
     * Do nothing; rendering hints not supported.
     */
    @Override
    public void setRenderingHints(Map<?, ?> arg0)
    {
        // Do nothing
    }

    
    /**
     * Set stroke if instance of BasicStroke, otherwise throw UnsupportedOperationException.
     * 
     * @param   stroke  stroke to use (must be BasicStroke)
     */
    @Override
    public void setStroke(Stroke stroke)
    {
        if (stroke instanceof BasicStroke)
        {
            this.stroke = (BasicStroke)stroke;
        }
        else
        {
            throw new UnsupportedOperationException("Only BasicStroke supported by ODGGraphics");
        }
    }

    
    /**
     * Set current graphics transformation to that supplied.
     * 
     * @param   newTransform  new graphics transformation to be used.
     */
    @Override
    public void setTransform(AffineTransform newTransform)
    {
        javaTransform = new AffineTransform(newTransform);
    }

    
    /**
     * Apply the specified shearing to the current graphics transformation. Subsequent
     * graphics operations will have the new combined transformation applied.
     * 
     * @param   sx  amount to shear x coordinates
     * @param   sy  amount to shear y coordinates
     */
    @Override
    public void shear(double sx, double sy)
    {
        javaTransform.shear(sx, sy);
    }

    
    /**
     * Concatenate the supplied transformation to the current graphics transformation. Subsequent
     * graphics operations will have the new combined transformation applied.
     */
    @Override
    public void transform(AffineTransform transform)
    {
        javaTransform.concatenate(transform);
    }
    
    
    /**
     * Translates the origin of the graphics context to the point
     * (<i>x</i>,&nbsp;<i>y</i>) in the current coordinate system.
     * Modifies this graphics context so that its new origin corresponds
     * to the point (<i>x</i>,&nbsp;<i>y</i>) in this graphics context's
     * original coordinate system.  All coordinates used in subsequent
     * rendering operations on this graphics context will be relative
     * to this new origin.
     * @param  x   the <i>x</i> coordinate.
     * @param  y   the <i>y</i> coordinate.
     */
    public void translate(int x, int y)
    {
        translate((double)x, (double)y);
    }

    
    /**
     * Translates the origin of the graphics context to the point
     * (<i>x</i>,&nbsp;<i>y</i>) in the current coordinate system.
     * Modifies this graphics context so that its new origin corresponds
     * to the point (<i>x</i>,&nbsp;<i>y</i>) in this graphics context's
     * original coordinate system.  All coordinates used in subsequent
     * rendering operations on this graphics context will be relative
     * to this new origin.
     * @param  x   the <i>x</i> coordinate.
     * @param  y   the <i>y</i> coordinate.
     */
    @Override
    public void translate(double x, double y)
    {
        javaTransform.translate(x, y);
    }
    
    
    
    /**
     * Utility method to map a set of vectors using an Affine transformation; basically ignores translation.
     * @param   m         affine transformation to use
     * @param   vectors   array [x1, y1, x2, y2, ..., xn, yn] of (x,y) coordinates of vectors to be mapped
     */
    private void mapVectors(AffineTransform m, double[] vectors)
    {
     // create a version of the transform without the translation
        double[] matrixCoeffs = {0,0,0,0,0,0};
        m.getMatrix(matrixCoeffs);
        // zero out translation components
        matrixCoeffs[4] = 0;
        matrixCoeffs[5] = 0;
        AffineTransform l = new AffineTransform(matrixCoeffs);
        
        l.transform(vectors, 0, vectors, 0, vectors.length/2);
    }
    
}
