/*
 * JODG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jodg;


import java.math.BigDecimal;
import java.math.RoundingMode;

import com.jsevy.jxml.XMLElement;

/**
 * The base class for all ODG entities. Provides some utility classes for scaling and the like.
 * 
 * @author jsevy
 *
 */
class ODGObject extends XMLElement
{
    private static int digits = 10;
    
    static String units = "in";
    static double scaleFactor = 72;
    
    
    public ODGObject()
    {
        super();
    }
    
    
    
    /**
     * Set the number of digits when a quantity is truncated with a call to setPrecision()
     * 
     * @param decimalDigits  Number of digits following the decimal point, between 0 and 16
     */
    public static void setPrecisionDigits(int decimalDigits)
    {
        digits = decimalDigits;
        
        // put in the range of 0 to 16
        if (digits < 0)
            digits = 0;
        if (digits > 16)
            digits = 16;
        
    }
    
    
    /**
     * Utility method used by subclasses to truncate output of decimal quantities
     * to the number of digits specified through a call to setPrecisionDigits; default is 10.
     * 
     * @param value  Double value to be truncated
     * @return  Input with digits following the decimal point truncated to the number of digits specified through a call to setPrecisionDigits
     */
    protected static double setPrecision(double value)
    {
        return (new BigDecimal(value)).setScale(digits, RoundingMode.HALF_UP).doubleValue();
    }
    
    
    /**
     * Given a double value, scale it according to the scale factor previously set, and add the units string.
     * @param value     Value to be scaled
     * @return          String giving the value with units appended.
     */
    protected static String scaleValueWithUnits(double value)
    {
        return Double.toString(value/scaleFactor) + units;
    }
    
    
    /**
     * Given a double value, scale it according to the scale factor previously set.
     * @param value     Value to be scaled
     * @return          Scaled value
     */
   protected static double scaleValue(double value)
    {
        return value/scaleFactor;
    }
    
    
    
    
    
}